import numpy as np
from scipy import sparse
import networkx as nx
import pandas as pd


def read_edge_list(edge_file, weight_col="weight", sep=","):
    edge_table = pd.read_csv(edge_file)
    edges = edge_table[["src", "trg"]].values
    node_labels, edges = np.unique(edges, return_inverse=True)
    id2label = dict(zip(np.arange(len(node_labels)), node_labels))
    edges = edges.reshape((edge_table.shape[0], 2))
    src = edges[:, 0]
    trg = edges[:, 1]

    if weight_col in edge_table.keys():
        w = edge_table["weight"].values
    else:
        w = np.ones_like(src)
    N = np.maximum(np.max(src), np.max(trg)) + 1
    A = sparse.csr_matrix((w, (src, trg)), shape=(N, N))
    return A, id2label


def to_adjacency_matrix(net, node_labels=None):
    if sparse.issparse(net):
        return sparse.csr_matrix(net), np.arange(net.shape[0])
    elif "networkx" in "%s" % type(net):
        if node_labels is not None:
            G = nx.subgraph(net, node_labels)
            return nx.adjacency_matrix(G, node_labels), node_labels
        else:
            return nx.adjacency_matrix(net), list(net.nodes())
    elif "numpy.ndarray" == type(net):
        return sparse.csr_matrix(net), np.arange(net.shape[0])


def to_community_matrix(community_ids):
    K = np.max(community_ids) + 1
    N = community_ids.size
    return sparse.csc_matrix((np.ones(N), (np.arange(N), community_ids)), shape=(N, K))


def construct_adjacency_matrix(src, dst, w, N):
    """
    Construct adjacency matrix from the list of edges

    Parameters
    ----------
    src : numpy.ndarray
        Source nodes
    dst : numpy.ndarray
        Target nodes
    w : numpy.ndarray
        Weight of edges
    N : int
        Number of nodes

    Returns
    -------
    A : scipy sparse matrix
        Adjacency matrix where A[i,j] indicate the
        weight of the edge from node i to node j
    """
    return sparse.csr_matrix((w, (src, dst)), shape=(N, N))


def find_non_self_loop_edges(A):
    """
    Parameters
    ---------
    A : scipy sparse matrix

    Returns
    ------
    r: numpy.ndarray
        Row ids for the non-zero elements
    c: numpy.ndarray
        Column ids for the non-zero elements
    v: numpy.ndarray
        Values of non-zero elements
    """
    r, c, v = sparse.find(A)
    non_self_loop = r != c
    r, c, v = r[non_self_loop], c[non_self_loop], v[non_self_loop]
    return r, c, v

    # Pack the community ids to an array
    community_ids = []
    if "networkx" in "%s" % type(net):
        community_ids = np.array([community_membership[x] for x in net.nodes()])
    else:
        community_ids = np.array([community_membership[x] for x in range(A.shape[0])])


def to_networkx_graph(A, nodes, create_using=nx.DiGraph):
    """
    Parameters
    ----------
    A : scipy sparse matrix
        Adjacency matrix
    nodes : list or numpy array
        List of node names in order of nodes in A

    Returns
    ------
    G : networkx.Graph
    """
    G = nx.from_scipy_sparse_matrix(A, create_using=create_using)
    return nx.relabel_nodes(G, dict(zip(np.arange(len(nodes)), nodes)))


def pairing(k1, k2, ordered=True):
    """
    Cantor pairing function
    """
    k12 = k1 + k2
    if ordered == False:
        return (k12 * (k12 + 1)) * 0.5 + np.minimum(k1, k2)
    else:
        return (k12 * (k12 + 1)) * 0.5 + k2


def depairing(z):
    """
    Inverse of Cantor pairing function
    """
    w = np.floor((np.sqrt(8 * z + 1) - 1) * 0.5)
    t = (w ** 2 + w) * 0.5
    y = np.round(z - t).astype(np.int64)
    x = np.round(w - y).astype(np.int64)
    return x, y



import numpy as np
from scipy import sparse
import pandas as pd
import networkx as nx
from cidre import utils
from cidre import filters


class Group:
    """Data class for anomalous node groups

    A node group consists of donors and recipients.
    Donors excessively provide edges to other nodes in the group.
    Recipients excessively receive edges from nodes in the group.

    The level of donorness and recipientness are measured by
    donor and recipient scores provided by an external algorithm.

    Then, donor and recipient nodes are identified by a threshold.
    """

    def __init__(self, node_ids, donor_scores, recipient_scores, threshold):
        """Initialize the group

        :param node_ids: ids of nodes
        :type node_ids: numpy.array
        :param donor_scores: donor scores.
        :type donor_scores: numpy.array
        :param recipient_scores: recipient scores
        :type recipient_scores: numpy.array
        :param threshold: threshold for classifying nodes into donors and recipients
        :type threshold: float
        """
        self.node_ids = node_ids.copy()
        self.recipient_score = recipient_scores.copy()
        self.donor_score = donor_scores.copy()
        self.threshold = threshold

        is_donor, is_recipient = (
            donor_scores >= threshold,
            recipient_scores >= threshold,
        )
        donors, recipients = node_ids[is_donor], node_ids[is_recipient]
        donor_scores, recipient_scores = (
            donor_scores[is_donor],
            recipient_scores[is_recipient],
        )
        self.donors = dict(zip(donors, donor_scores))
        self.recipients = dict(zip(recipients, recipient_scores))

    def size(self):
        return len(self.node_ids)

    def set_within_net(self, A, node_ids):
        brow = A[:, node_ids].sum(axis=0)
        bcol = A[node_ids, :].sum(axis=1)
        As = A[:, node_ids][node_ids, :].toarray()

        self.num_within_edges = As.sum() - As.diagonal().sum()

        # Add 'Other' node to the adjacency matrix
        B = np.block([[As, bcol], [brow, np.array([0])]])
        self.A = np.array(B)

    def get_within_net(self):
        return self.A

    def get_num_edges(self):
        return self.num_within_edges

    def get_donors(self):
        return self.donors

    def get_recipients(self):
        return self.recipients

    def get_donor_recipients(self):
        ids = set(list(self.donors.keys())).intersection(
            set(list(self.recipients.keys()))
        )
        return {
            i: {"donor": self.donors[i], "recipient": self.recipients[i]} for i in ids
        }


class Cidre:
    def __init__(
        self,
        min_edge_weight=0,
        min_expected_weight=1,
        group_membership=None,
        alpha=0.01,
    ):
        if isinstance(group_membership, list):
            group_membership = np.array(group_membership)

        self.group_membership = group_membership
        self.edge_filter = filters.EdgeFilter(
            alpha=alpha,
            min_edge_weight=min_edge_weight,
            min_expected_weight=min_expected_weight,
            remove_selfloop=True,
        )

    def detect(self, G, threshold):
        """Detecting anomalous groups with excessive donors and recipients.

        :param G: Network. If G is a sparse matrix, entry G[i,j] should be the weight of the edge from node i to j.
        :type G: scipy sparse matrix or networkx.Graph.
        :param threshold: Threshold for donor and recipient scores. A higher thhreshold yields smaller and tighter groups
        :type threshold: float
        """
        #
        # Input parsing
        #
        A, node_labels = utils.to_adjacency_matrix(G)
        #
        # Edge filtering
        #
        self.edge_filter.fit(A, group_membership=self.group_membership)
        src, dst, w = sparse.find(A)
        src, dst, w = self.edge_filter.transform(src, dst, w)

        Abar = utils.construct_adjacency_matrix(
            src, dst, w, A.shape[0]
        )  # Edge filtered network

        #
        # Node filtering
        #
        num_nodes = A.shape[0]  # number of nodes
        U = np.ones(num_nodes)  # nodes not truncated
        indeg = np.maximum(
            np.array(A.sum(axis=0)).ravel(), 1.0
        )  # In-degree. Clip to one to avoid zero divide.
        outdeg = np.maximum(
            np.array(A.sum(axis=1)).ravel(), 1.0
        )  # Out-degree. Clip to one to avoid zero divide.
        while True:
            # Compute the donor score, recipient score and cartel score
            donor_score = np.multiply(U, (Abar @ U) / outdeg)
            recipient_score = np.multiply(U, (U @ Abar) / indeg)

            # Drop the nodes with a cartel score < threshold
            drop_from_U = (U > 0) * (
                np.maximum(donor_score, recipient_score) < threshold
            )

            # Break the loop if no node is dropped from the cartel
            if np.any(drop_from_U) == False:
                break

            # Otherwise, drop the nodes from the cartel
            U[drop_from_U] = 0

        # Find the nodes in U
        survived_nodes = np.where(U)[0]
        Abar = Abar[:, survived_nodes][survived_nodes, :].copy()

        # Partition U into disjoint groups, U_l
        anomalous_group_list = []
        Gbar = nx.from_scipy_sparse_matrix(Abar, create_using=nx.DiGraph)
        Gbar.remove_nodes_from(list(nx.isolates(Gbar)))
        for _, _nd in enumerate(nx.weakly_connected_components(Gbar)):
            _nodes = survived_nodes[np.array(list(_nd))]

            # Remove the group U_l if
            # U_l does not contain edges less than or equal to
            # min_group_edge_num
            # A_Ul = A[_nodes, :][:, _nodes]

            group = Group(
                _nodes, donor_score[_nodes], recipient_score[_nodes], threshold
            )
            group.set_within_net(A, _nodes)
            anomalous_group_list += [group]

        return anomalous_group_list



import pandas as pd
import numpy as np
from scipy import sparse
import seaborn as sns
import matplotlib.pyplot as plt
from matplotlib import patches
import matplotlib.colors as colors
import textwrap
import re


class DrawCartel:
    def __init__(self):
        self.theta_1 = np.pi * 0.7
        self.angle_margin = 3 * np.pi / 25
        self.node2color = {}
        self.angles = None
        self.radius = 2
        self.label_node_margin = 0.35
        self.group_arc_node_margin = 0.1
        self.edge_norm = lambda x: np.power(x, 1 / 2)
        self.max_edge_width = 15
        self.font_size = 15
        self.node_size = 0.25
        self.label_width = 15
        self.max_label_width = 40
        self.group_order = {"source": 0, "target": 1, "reciprocal": 2, "other": 3}

    def draw(
        self, group, node_labels=None, cmap=None, ax=None,
    ):
        """
        Draw citation networks within a citation cartel

        Parameters
        ----------
        A : scipy.sparse matrix
            The adjacency matrix for the network with all nodes
        node_ids : np.array or list
            The node ids of the nodes in the cartel
        donor_score : np.array or list
            The donor score of the nodes in the cartel.
            We assume the donor_score[i] indicates the
            donor score for node_ids[i]
        recipient_score : np.array or list
            The recipient score of the nodes in the cartel.
            We assume the recipient_score[i] indicates the
            recipient score for node_ids[i]
        theta : float
            The threshold for the donor and recipient score
        node_name : list
            Node names. Assume that node_name[i] indicates the
            name of the node_ids[i]
        cmap : matplotlib color map or list
            Color map or List of strings indicating hex code
        ax : axis

        Return
        ------
        ax : axis
        """

        #
        # Input formatting
        #
        # A,
        # node_ids,
        # donor_score,
        # recipient_score,
        # theta,
        node_ids = np.array(group.node_ids)
        donor_score = np.array(group.donor_score)
        recipient_score = np.array(group.recipient_score)
        theta = group.threshold

        if node_labels is None:
            node_names = ["%d" % i for i in node_ids]
        else:
            node_names = [node_labels[i] for i in node_ids]

        # Classify nodes into donor, recipient and reciprocal
        node_types = self.classify_nodes(donor_score, recipient_score, theta)

        # Change the angle for the reciprocal journals based on the number of it
        num_reciprocal = np.sum(
            np.array([node_types[i] == "reciprocal" for i in range(len(node_types))])
        )
        if num_reciprocal > 2:
            self.theta_1 = np.pi * 0.3
        else:
            self.theta_1 = np.pi * 0.7

        #
        # Construct the adjacency matrix with 'Other' node
        #
        As = group.A.copy()
        node_types += ["other"]
        node_names += ["Other"]

        #
        # Calculate the positions and sizes
        #

        # Make node table
        num_nodes = len(node_types)
        node_table = pd.DataFrame(
            {"id": np.arange(num_nodes), "group": node_types, "name": node_names}
        )

        # Calculate the angle of each node
        node_table, self.angles = self.calc_node_angles(node_table)

        # Calculate the position of journals based on the angles
        node_table = self.calc_node_position(node_table)

        # Text folding
        node_table = self.fold_node_name(node_table)

        # Compute the edge positions based on the nodes
        edge_table = self.make_edge_table(node_table, As)

        # make color map
        self.make_color_map(node_table, cmap)

        #
        # Plot
        #
        self.plot_edges(node_table, edge_table, ax)

        self.plot_node_label(node_table, ax)

        self.plot_nodes(node_table, As, node_ids, ax)

        self.plot_group_arc(node_table, ax)

        self.trim(ax)

        return ax

    def classify_nodes(self, donor_score, recipient_score, threshold):
        is_recipient = recipient_score >= threshold
        is_donor = donor_score >= threshold
        is_reciprocal = is_recipient & is_donor
        is_recipient = is_recipient & (~is_reciprocal)
        is_donor = is_donor & (~is_reciprocal)
        node_type = is_recipient + 2 * is_donor + 3 * is_reciprocal
        node_type = np.array(["", "target", "source", "reciprocal"])[node_type]
        return node_type.tolist()

    def calc_node_angles(self, node_table):
        # Compute the coordinate of nodes
        self.theta_2 = np.pi - self.theta_1 - 2 * self.angle_margin

        node_table["within_group_id"] = -1
        node_table["angle"] = -1
        angles = {"margin_angle": self.angle_margin}
        for group_name in ["source", "target", "reciprocal", "other"]:
            dg = node_table[node_table.group == group_name]
            if group_name == "source":
                start_angle = -self.angle_margin - self.theta_1 - self.theta_2 / 2
                end_angle = start_angle + self.theta_1
            elif group_name == "target":
                start_angle = self.angle_margin + self.theta_2 / 2
                end_angle = start_angle + self.theta_1
            elif group_name == "reciprocal":
                start_angle = -self.theta_2 / 2
                end_angle = start_angle + self.theta_2
            elif group_name == "other":
                start_angle = self.theta_2 / 2 + self.angle_margin * 2 + self.theta_1
                end_angle = start_angle + self.theta_2

            ids = np.arange(dg.shape[0])
            node_table.loc[dg.index, "within_group_id"] = ids
            n = dg.shape[0]

            if (group_name == "reciprocal") and (n >= 2):
                a = (
                    (ids) * ((end_angle - start_angle) - self.angle_margin) / (n - 1)
                    + start_angle
                    + 0.5 * self.angle_margin
                )
            else:
                if n >= 2:
                    a = (
                        ids
                        * ((end_angle - start_angle) - 1.5 * self.angle_margin)
                        / (n - 1)
                        + start_angle
                        + 0.75 * self.angle_margin
                    )
                else:
                    a = (
                        (ids + 1)
                        * ((end_angle - start_angle) - 1.5 * self.angle_margin)
                        / (n + 1)
                        + start_angle
                        + 0.75 * self.angle_margin
                    )
                # node_table.loc[dg.index, "angle"] = (ids +1) * angle_group / (n-1) + start_angle
            node_table.loc[dg.index, "angle"] = a

            angles[group_name] = {"start": start_angle, "end": end_angle}
        return node_table, angles

    def calc_node_position(self, node_table):
        nodes = node_table.copy()
        nodes["x"] = self.radius * np.sin(nodes.angle)
        nodes["y"] = self.radius * np.cos(nodes.angle)
        return nodes

    def make_edge_table(self, node_table, As):

        # Compute the edge table
        src, trg = np.where(As)
        selfloop = src != trg
        src, trg = src[selfloop], trg[selfloop]
        w = As[(src, trg)]
        edge_table = pd.DataFrame({"src": src, "trg": trg, "w": w})

        edges = edge_table.copy()
        edges = pd.merge(
            edges,
            node_table[["id", "x", "y"]],
            left_on="src",
            right_on="id",
            how="left",
        ).rename(columns={"x": "src_x", "y": "src_y"})
        edges = pd.merge(
            edges,
            node_table[["id", "x", "y"]],
            left_on="trg",
            right_on="id",
            how="left",
        ).rename(columns={"x": "trg_x", "y": "trg_y"})

        # Normalize the maximum to be one
        wmax = np.maximum(
            np.max(np.triu(As[:, :-1][:-1, :], 1)),
            np.max(np.tril(As[:, :-1][:-1, :], 1)),
        )
        edges["w"] = edges["w"] / wmax
        edges["w"] = self.edge_norm(edges["w"])
        edges["w"] = edges["w"] / edges["w"].max()
        n = As.shape[0] - 1
        return edges

    def make_color_map(self, node_table, cmap):
        n = node_table.shape[0]

        # sort nodes
        _node_table = node_table.copy()
        _node_table["group_order"] = _node_table.apply(
            lambda x: self.group_order[x["group"]], axis=1
        )
        _node_table = _node_table.sort_values(by="group_order")

        if cmap is None:
            if _node_table.shape[0] <= 8:
                cmap = sns.color_palette().as_hex()
            elif _node_table.shape[0] <= 20:
                cmap = sns.color_palette().as_hex()
                cmap2 = sns.color_palette("husl", 12).as_hex()
                cmap = cmap + cmap2
            elif _node_table.shape[0] <= 20:
                cmap = sns.color_palette().as_hex()
                cmap2 = sns.color_palette("Paired").as_hex()
                cmap = (
                    cmap
                    + [c for i, c in enumerate(cmap2) if i % 2 == 1]
                    + [c for i, c in enumerate(cmap2) if i % 2 == 0]
                )
            elif _node_table.shape[0] <= 40:
                # cmap = sns.color_palette("Set1").as_hex()
                cmap = sns.color_palette("tab20").as_hex()
                cmap_list = []
                for l in range(5):
                    cmap_list += [c for i, c in enumerate(cmap) if i % 2 == l]
                cmap_1 = cmap_list
                cmap = sns.color_palette("tab20b").as_hex()
                cmap_list = []
                for l in range(5):
                    cmap_list += [c for i, c in enumerate(cmap) if i % 2 == l]
                cmap_2 = cmap_list
                cmap = cmap_1 + cmap_2
            else:
                cmap = sns.color_palette("Spectral", n).as_hex()

        self.node2color = {}
        i = 0

        for _, row in _node_table.iterrows():
            if row["group"] == "other":
                self.node2color[row["id"]] = "#c4c4c4"
            else:
                self.node2color[row["id"]] = cmap[i]
                i += 1

        return self.node2color

    def plot_edges(self, node_table, edge_table, ax):
        _edges = edge_table.sort_values(by="w").sort_values(
            by=["src", "trg"], ascending=False
        )
        for i, edge in _edges.iterrows():
            if edge.src == edge.trg:
                continue
            x_pos = edge.src_x
            y_pos = edge.src_y
            dx = edge.trg_x - edge.src_x
            dy = edge.trg_y - edge.src_y
            length = np.sqrt(dx * dx + dy * dy)
            orient = np.array([edge.trg_x, edge.trg_y])

            w = edge["w"]
            style = """Simple,tail_width={w},head_width={w1}
                  """.format(
                w=w * self.max_edge_width, w1=2 * w * self.max_edge_width,
            )

            src_id = int(edge.src)
            trg_id = int(edge.trg)
            color = self.node2color[int(edge.src)]
            color = (
                color + "66"
                if node_table.iloc[src_id]["group"] == "other"
                else color + "cc"
            )
            kw = dict(arrowstyle=style, color=color, linewidth=0,)
            connectionstyle = "arc3,rad=.2"
            a3 = patches.FancyArrowPatch(
                (edge.src_x, edge.src_y),
                (edge.trg_x, edge.trg_y),
                # shrinkB=15,
                connectionstyle=connectionstyle,
                **kw
            )
            ax.add_patch(a3)

    def plot_node_label(self, node_table, ax):
        reciprocal_num = np.sum(node_table["group"].values == "reciprocal")

        texts = []
        for i, row in node_table.iterrows():
            x = row["x"] + (self.label_node_margin) * np.sin(row["angle"])
            y = row["y"] + (self.label_node_margin) * np.cos(row["angle"])
            if (x > 0) and (y > 0):
                ha = "left"
                va = "bottom"
            elif (x > 0) and (y < 0):
                ha = "left"
                va = "top"
            elif (x < 0) and (y < 0):
                ha = "right"
                va = "top"
            elif (x < 0) and (y > 0):
                ha = "right"
                va = "bottom"

            va = "center"
            if row["group"] == "reciprocal":
                if reciprocal_num > 2:
                    pass
                else:
                    ha = "center"
                    va = "bottom"
            if row["group"] == "other":
                ha = "center"
                va = "top"
            texts += [ax.text(x, y, row["name"], ha=ha, va=va, fontsize=self.font_size)]

    def plot_group_arc(self, node_table, ax):
        params = {"lw": 3, "fc": "white", "alpha": 0.3, "zorder": 0, "angle": 90}

        a3 = patches.Arc(
            (0, 0),
            2 * self.radius * (self.group_arc_node_margin + 1),
            2 * self.radius * (self.group_arc_node_margin + 1),
            theta1=-180 * self.angles["reciprocal"]["end"] / np.pi,
            theta2=-180 * self.angles["source"]["start"] / np.pi,
            ec="black",
            ls="-",
            **params
        )
        ax.add_patch(a3)

        a3 = patches.Arc(
            (0, 0),
            2 * self.radius * (self.group_arc_node_margin + 1.05),
            2 * self.radius * (self.group_arc_node_margin + 1.05),
            theta2=-180 * self.angles["reciprocal"]["start"] / np.pi,
            theta1=-180 * self.angles["target"]["end"] / np.pi,
            ec="black",
            ls="--",
            **params
        )
        ax.add_patch(a3)

    def trim(self, ax):
        R = self.radius * 1.2
        ax.set_xlim(left=-R, right=R)
        ax.set_ylim(bottom=-R, top=R)
        ax.axis("off")

    def plot_nodes(self, node_table, As, node_ids, ax):

        # Calculate the angle of pie
        indeg = np.array(As.sum(axis=0)).reshape(-1)[:-1]
        share = As[:-1, :-1] @ np.diag(1.0 / np.maximum(1, indeg))

        for i, row in node_table.iterrows():
            if row["group"] == "other":
                ax.pie(
                    [1],
                    startangle=90,
                    colors=[self.node2color[row["id"]]],
                    center=(row["x"], row["y"]),
                    radius=self.node_size,
                    wedgeprops={"edgecolor": self.node2color[i], "linewidth": 3},
                )
            else:
                order = np.argsort(share[:, i])

                node_color_list = [self.node2color[j] for j in order]

                ax.pie(
                    [1 - np.sum(share[:, i])]
                    + np.array(share[order, i]).reshape(-1).tolist(),
                    startangle=90,
                    colors=["#ffffffff"] + node_color_list,
                    center=(row["x"], row["y"]),
                    radius=self.node_size,
                    wedgeprops={"linewidth": 0},
                )
                c = patches.Circle(
                    (row["x"], row["y"]),
                    self.node_size,
                    fill=None,
                    edgecolor=self.node2color[row["id"]],
                    linewidth=3,
                )
                ax.add_patch(c)

    def fold_node_name(self, node_table):
        for g, dg in node_table.groupby("group"):
            if dg.shape[0] < 8:
                node_table.loc[dg.index, "name"] = dg["name"].apply(
                    lambda x: self.fold_text(
                        x, width=self.label_width, max_width=self.max_label_width
                    )
                )
            else:
                node_table.loc[dg.index, "name"] = dg["name"].apply(
                    lambda x: self.fold_text(x, width=45, max_width=99999)
                )
        return node_table

    def fold_text(self, txt, width, max_width):
        if width == "auto":
            w = 10
            while True:
                s = textwrap.wrap(txt, w)
                if (len(s) <= 2) or (w >= max_width):
                    break
                w += 1
            txt = "\n".join(s)
        elif width is not None:
            txt = "\n".join(textwrap.wrap(txt, width))
        return txt



"""Edge filtering function"""
import numpy as np
from scipy import stats, sparse
import sys
from cidre import utils
from functools import partial


class EdgeFilter:
    def __init__(
        self, min_edge_weight=0, alpha=0.01, remove_selfloop=True, min_expected_weight=1
    ):
        self.alpha = alpha
        self.min_edge_weight = min_edge_weight
        self.remove_selfloop = remove_selfloop
        self.min_expected_weight = min_expected_weight

    def fit(self, A, group_membership, mask=None):
        """Find the excessive edges in the network using the dcSBM as a null model.

        :param A: Network
        :type A: scipy sparse matrix
        :param group_membership: group membership of nodes. group_membership[i] is the group to which node i belongs.
        :type group_membership: numpy.array
        :param mask: mask[i,j] = 1 to set edge (i,j) to be insignificant, defaults to None
        :type mask: scipy sparse matrix, optional
        """
        if group_membership is None:
            group_membership = np.zeros(A.shape[0]).astype(int)
        p_value, src, dst, w = self._calc_p_values_dcsbm(
            A, group_membership, self.min_expected_weight
        )

        if self.remove_selfloop:
            s = src != dst
            p_value, src, dst, w = p_value[s], src[s], dst[s], w[s]

        # Remove edges less than minimum weight
        if self.min_edge_weight > 0:
            s = w >= self.min_edge_weight
            p_value, src, dst, w = p_value[s], src[s], dst[s], w[s]

        # Mask pre-selected edges
        if mask is not None:
            wth = np.array(mask[(src_, trg_)]).reshape(-1)
            s = np.isclose(wth, 0)
            p_value, src, dst, w = p_value[s], src[s], dst[s], w[s]

        # Perform the Benjamini-Hochberg statistical test
        is_significant = self._benjamini_hochberg_test(p_value)

        # Find the excessive edges
        src, dst, w = src[is_significant], dst[is_significant], w[is_significant]

        # Construct the filter
        self.filter = self._make_filter_func(src, dst, None, A.shape[0])

    def transform(self, src, dst, w):
        return self.filter(src, dst, w)

    def _calc_p_values_dcsbm(self, A, group_membership, min_expected_weight=1):
        """Calculate the p_values using the degree-corrected stochastic block model.

        :param A: Adjacency matrix. Adjacency matrix, where A[i,j] indicates the weight of the edge from node i to node j.
        :type A: scipy sparse matrix
        :param group_membership: group_membership[i] indicates the ID of the group to which node i belongs
        :type group_membership: numpy.array
        :return: p-values
        :rtype: float
        """

        N = A.shape[0]
        indeg = np.array(A.sum(axis=0)).reshape(-1)
        outdeg = np.array(A.sum(axis=1)).reshape(-1)
        C_SBM = utils.to_community_matrix(group_membership)

        Lambda = C_SBM.T @ A @ C_SBM
        Din = np.array(Lambda.sum(axis=0)).reshape(-1)
        Dout = np.array(Lambda.sum(axis=1)).reshape(-1)

        theta_in = indeg / np.maximum(C_SBM @ Din, 1.0)
        theta_out = outdeg / np.maximum(C_SBM @ Dout, 1.0)

        src, dst, w = utils.find_non_self_loop_edges(A)
        lam = (
            np.array(Lambda[group_membership[src], group_membership[dst]]).reshape(-1)
            * theta_out[src]
            * theta_in[dst]
        )
        lam = np.maximum(lam, min_expected_weight)
        pvals = 1.0 - stats.poisson.cdf(w - 1, lam)

        return pvals, src, dst, w

    def _benjamini_hochberg_test(self, pvals):
        """Benjamini-Hochberg statistical test

        :param pvals: p-values
        :type pvals: numpy.array
        :return: significant[i] = True if the ith element is significant. Otherwise signfiicant[i] = False.
        :rtype: numpy.array (bool)
        """
        order = np.argsort(pvals)
        M = pvals.size
        is_sig = pvals[order] <= (self.alpha * np.arange(1, M + 1) / M)
        if np.any(is_sig) == False:
            return is_sig

        last_true_id = np.where(is_sig)[0][-1]
        is_sig = np.zeros(M)
        is_sig[order[: (last_true_id + 1)]] = 1
        return is_sig > 0

    def _make_filter_func(self, src, trg, wth, N):
        """Make a filter function

        :param src: Source node
        :type src: np.array
        :param trg: Target node
        :type trg: np.array
        :param wth: Minimum Weight of edges between source and target nodes
        :type wth: numpy.array
        :param N: Number of nodes
        :type N: int
        :return: Filtering function
        :rtype: function
        """
        if wth is None:
            wth = 1e-8 * np.ones_like(src)

        # Convert pairs of integers into integers
        W = sparse.csr_matrix((wth, (src, trg)), shape=(N, N))

        # Define the is_excessive function for CIDRE
        def is_excessive(src_, trg_, w_, W):
            wth = np.array(W[(src_, trg_)]).reshape(-1)
            s = (w_ >= wth) * (wth > 0)
            return src_[s], trg_[s], w_[s]

        return partial(is_excessive, W=W)



#!/usr/bin/env python
# encoding: utf-8
import os
from os import path

from setuptools import find_packages, setup

__version__ = "0.0.0"


def load_requires_from_file(fname):
    if not os.path.exists(fname):
        raise IOError(fname)
    return [pkg.strip() for pkg in open(fname, "r")]


this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, "README.md")) as f:
    long_description = f.read()

setup(
    name="cidre",
    version=__version__,
    author="Sadamori Kojaku",
    url="https://github.com/skojaku/cidre",
    description="Algorithm for finding anomalous groups in networks",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=find_packages(),
    install_requires=load_requires_from_file("requirements.txt"),
    include_package_data=True,
    zip_safe=False,
    license="MIT",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Science/Research",
        "Topic :: Software Development",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    keywords="networks, community detection, anomaly detection",
)
